/* "comm" match extension for Xtables */

#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/sched.h>
#include <linux/file.h>
#include <linux/fdtable.h>
#include <linux/version.h>
#include <net/sock.h>
#include <net/inet_sock.h>
#include <net/tcp.h>
#include <net/udp.h>
#include <net/inet_hashtables.h>
#include <linux/rcupdate.h>

#include <linux/netfilter_ipv4.h>
#include <linux/netfilter/x_tables.h>

#include "xt_comm.h"

MODULE_LICENSE("GPL");
MODULE_AUTHOR("mengyuan");
MODULE_DESCRIPTION("xtables module for comm match");
MODULE_ALIAS("ip6t_comm");
MODULE_ALIAS("ipt_comm");

static bool comm_mt(const struct sk_buff *skb, struct xt_action_param *par)
{
	const struct xt_comm_info *info = par->matchinfo;
	struct task_struct *p;
	struct files_struct *files;
	struct fdtable *fdt;
	struct file *file = NULL;
	struct socket *sock;
	struct socket *lsock = NULL;
	int i, err;
	bool ret = false;
	const struct tcphdr *th;
	struct tcphdr _tcph;
	const struct udphdr *uh;
	struct udphdr _udph;
	struct sock *sk;
	const struct iphdr *iph = ip_hdr(skb);
	int sdif = inet_sdif(skb);
	struct net *net = dev_net(skb_dst(skb)->dev);

	/* LOCAL_OUT always have socket file,
	 *  no need to lookup listener hash table */
	if (par->fragoff != 0 || par->state->out)
		goto task;
	/* UDP INPUT */
	if (IPPROTO_UDP == iph->protocol) {
		uh = skb_header_pointer(skb, par->thoff, sizeof(_udph), &_udph);
		if (uh == NULL)
			goto task;
		sk = __udp4_lib_lookup(net, iph->saddr, uh->source, iph->daddr,
				uh->dest, inet_iif(skb), sdif, &udp_table, NULL);
		if (sk) {
			lsock = sk->sk_socket;
			goto task;
		}
	}
	/* TCP INPUT */
	if (IPPROTO_TCP == iph->protocol) {
		th = skb_header_pointer(skb, par->thoff, sizeof(_tcph), &_tcph);
		if (th == NULL)
			goto task;
		sk = __inet_lookup_listener(net, &tcp_hashinfo, skb, __tcp_hdrlen(th),
			iph->saddr, th->source, iph->daddr, ntohs(th->dest), inet_iif(skb), sdif);
		if (sk)
			lsock = sk->sk_socket;
	}
task:
	/* rcu_read_lock()ed by nf_hook */
	list_for_each_entry_rcu(p, &init_task.tasks, tasks) {
		if (strncmp(p->comm, info->comm, sizeof(p->comm)))
			continue;
		task_lock(p);
		files = p->files;
		if (files) {
			rcu_read_lock();
			fdt = files_fdtable(files);
			for (i = 0; i < fdt->max_fds; i++) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 11, 0)
				file = files_lookup_fd_locked(files, i);
#else
				file = fcheck_files(files, i);
#endif
				if (!file)
					continue;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 11, 0)
				sock = sock_from_file(file);
#else
				sock = sock_from_file(file, &err);
#endif
				if (!sock)
					continue;
				if (sock == lsock) {
					ret = true;
					break;
				}
				if (!skb || !skb->sk || !skb->sk->sk_socket)
					continue;
				if (sock == skb->sk->sk_socket) {
					ret = true;
					break;
				}
			}
			rcu_read_unlock();
		}
		task_unlock(p);
	};

	return ret ^ info->invert;
}

static struct xt_match comm_mt_reg __read_mostly = {
	.name      = "comm",
	.revision  = 0,
	.family    = NFPROTO_UNSPEC,
	.match     = comm_mt,
	.matchsize = sizeof(struct xt_comm_info),
	.hooks     = (1 << NF_INET_LOCAL_IN) | (1 << NF_INET_LOCAL_OUT),
	.me        = THIS_MODULE,
};

static int __init comm_mt_init(void)
{
	return xt_register_match(&comm_mt_reg);
}

static void __exit comm_mt_exit(void)
{
	xt_unregister_match(&comm_mt_reg);
}

module_init(comm_mt_init);
module_exit(comm_mt_exit);
